// This file is a companion file of the paper
// Stretch Factor of Long Paths in a planar Poisson-Delaunay Triangulation
// by Nicolas Chenavier and Olivier Devillers.
//
// This file has to be compiled using CGAL Library (www.cgal.org)
// 
// (c) Olivier Devillers (INRIA) 2016







#include <fstream>
#include <set>

// CGAL headers
#include <CGAL/Exact_predicates_inexact_constructions_kernel.h>
#include <CGAL/Delaunay_triangulation_2.h>

#include <CGAL/Triangulation_vertex_base_with_info_2.h>
#include <CGAL/Random.h>




/* A vertex class with an additionnal handle */
template < class Gt, class Vb = CGAL::Triangulation_vertex_base_2<Gt> >
class My_vertex_base
  : public  Vb
{
  typedef Vb                              Base;
public:
  typedef typename Vb::Vertex_handle      Vertex_handle;
  typedef typename Vb::Face_handle        Face_handle;
  typedef typename Vb::Point              Point;
  template < typename TDS2 >
  struct Rebind_TDS {
    typedef typename Vb::template Rebind_TDS<TDS2>::Other    Vb2;
    typedef My_vertex_base<Gt,Vb2>                           Other;
  };
public:
  Vertex_handle  previousS;
  double length;
  Vertex_handle  previousU;
  Vertex_handle  previousG;
public:
  My_vertex_base() : Base() {}
  My_vertex_base(const Point & p) : Base(p) {}
  My_vertex_base(const Point & p, Face_handle f) : Base(f,p) {}
  My_vertex_base(Face_handle f) : Base(f) {}
};

typedef CGAL::Exact_predicates_inexact_constructions_kernel K;
typedef K::Point_2 Point_2;
typedef K::Iso_rectangle_2 Iso_rectangle_2;
typedef My_vertex_base<K> Vb;
typedef CGAL::Triangulation_face_base_2<K> Fb;
typedef CGAL::Triangulation_data_structure_2<Vb,Fb> Tds;
typedef CGAL::Delaunay_triangulation_2<K,Tds> Delaunay;

  class Comparator
  {
  public:
    bool operator()(const Delaunay::Vertex_handle &a1, const Delaunay::Vertex_handle  &a2)
      {
        return a1->length< a2->length ;
      }
  };




typedef  std::multiset<Delaunay::Vertex_handle,Comparator> Dijkstra_queue;






int main(int argc, char **argv)
{
  static CGAL::Random  rnd_gen;
  Delaunay dt; 

int N_points, N_trials;
// Pour la somme
double  Torigin_length=0;
double  Torigin_degree=0;
double  Tinternal_length=0;
double  Tabove_length=0;
double  Tbelow_length=0;
double  Tslope_length=0;
double  Tshortest_length=0;
double  Tmax=0;
double  Tinternal_edges=0;
double  Tabove_edges=0;
double  Tbelow_edges=0;
double  Tslope_edges=0;
double  Tshortest_edges=0;
// pour la somme des carrés
double  TTorigin_length=0;
double  TTorigin_degree=0;
double  TTinternal_length=0;
double  TTabove_length=0;
double  TTbelow_length=0;
double  TTslope_length=0;
double  TTshortest_length=0;
double  TTinternal_edges=0;
double  TTabove_edges=0;
double  TTbelow_edges=0;
double  TTslope_edges=0;
double  TTshortest_edges=0;

// for visualisation purpose, coordinates are in   pixel
double LENGTH=600;
double Xmax=330;
double Xmin=-Xmax;
 double Ymax= LENGTH*LENGTH/(Xmax-Xmin)/2;
double Ymin=-Ymax;
double Xsource=-300;
double Ysource=0; 
double Xtarget=300;
double Ytarget=Ysource;

  std::cout<<"Estimating the stretch ratio in a Delaunay triangulation"<<std::endl
	   <<"Usage : "<< argv[0]<<" number_of_points number_of_trials "<<std::endl;
  if ((argc!=3)) exit(0); 
  N_points=atoi(argv[1]);
  N_trials=atoi(argv[2]);




 for ( int N_it=0; N_it < N_trials ; ) {


    double  origin_length=0;
    double  origin_degree=0;
    double  internal_length=0;
    double  above_length=0;
    double  below_length=0;
    double  slope_length=0;
    double  shortest_length=0;
    double ymin=0,ymax=0;
    int  internal_edges=0;
    int  above_edges=0;
    int  below_edges=0;
    int  slope_edges=0;
    int  shortest_edges=0;

    std::vector<Delaunay::Point> points;
    points.reserve(N_points);
    for(int i=0; i<N_points; ++i) 
      //      points.push_back( Delaunay::Point( rnd_gen.get_double(-100,110), rnd_gen.get_double(-34,34) ) );
      points.push_back( Delaunay::Point( rnd_gen.get_double(Xmin,Xmax), rnd_gen.get_double(Ymin,Ymax) ) );
    // compute Delaunay
    dt.clear();
    Delaunay::Point debut(Xsource,Ysource), fin(Xtarget,Ytarget);
    Delaunay::Vertex_handle source= dt.insert(debut);
    Delaunay::Vertex_handle target= dt.insert(fin);
    dt.insert(points.begin(), points.end());
    //std::cout<<dt.number_of_vertices()<<" sites in [-0.1,1.1]x[-0.42,0.42]  trial nb "<<N_it<<std::flush<<std::endl;
    for( Delaunay::Vertex_iterator vit=dt.vertices_begin(); vit!=dt.vertices_end(); ++vit) {
      vit->previousS = Delaunay::Vertex_handle();
      vit->previousU = Delaunay::Vertex_handle();
      vit->previousG = Delaunay::Vertex_handle();
      vit->length = 1000000;
    }
    
    source->length=0;
    
    //computing first face in straight walk
    Delaunay::Face_circulator fc(source);
    int i;
    while(true){ 
      i=fc->index(source);
      //      std::cout<<fc->vertex(dt.ccw(i))->point().y() <<"  "<< fc->vertex(dt.cw(i))->point().y() <<std::endl;
      if (   ( fc->vertex(dt.ccw(i))->point().y()<Ysource )
	     && ( fc->vertex(dt.cw(i))->point().y() >Ysource ) ) break;
      ++fc;
    }
    Delaunay::Face_handle f=fc;
    
    Delaunay::Face_handle previousf;
    Delaunay::Vertex_handle newv;
    Delaunay::Vertex_handle top=f->vertex(dt.cw(i));
    Delaunay::Vertex_handle bot=f->vertex(dt.ccw(i));
    Delaunay::Vertex_handle previousv=source;
    Delaunay::Vertex_handle greedy=source;
    double slope_below;
    double slope_above;
    above_length=sqrt( CGAL::squared_distance(source->point(),top->point()));
    below_length=sqrt( CGAL::squared_distance(source->point(),bot->point()));
    above_edges=1;
    below_edges=1;
 
    // loop for straight walk
    while (true){
      previousf = f;
      f=f->neighbor( f->index(previousv) );
      newv=f->vertex( f->index(previousf) );
      internal_length += sqrt( CGAL::squared_distance(top->point(),bot->point()));
      ++internal_edges;

       if (previousv == greedy){ // minimal slope path increases
	slope_below = (bot->point().y() - greedy->point().y()) / (bot->point().x() - greedy->point().x());
	slope_above =  (top->point().y() - greedy->point().y()) / (top->point().x() - greedy->point().x());
	if ( (slope_above > 0)&&(slope_below<0))// usual case
	  if (slope_above<-slope_below) {
	    slope_length+=sqrt(CGAL::squared_distance(greedy->point(),top->point()));
	    greedy = top;
	  } else {
	    slope_length+=sqrt(CGAL::squared_distance(greedy->point(),bot->point()));
	    greedy = bot;
	  }
	else{
	  if(greedy->point().y()>Ysource){
	    slope_length+=sqrt(CGAL::squared_distance(greedy->point(),top->point()));
	    greedy = top;
	  } else {
	    slope_length+=sqrt(CGAL::squared_distance(greedy->point(),bot->point()));
	    greedy = bot;
	  }
	}
	++slope_edges;
      }

      if (newv == target){ // walk end
	above_length += sqrt( CGAL::squared_distance(top->point(),newv->point()));
	++above_edges;
	below_length += sqrt( CGAL::squared_distance(bot->point(),newv->point()));
	below_edges++;
	slope_length+=sqrt(CGAL::squared_distance(greedy->point(),target->point()));
	slope_edges++;
	break;
      }      

      // increase upper or lower path
      if ( newv->point().y() > Ysource) {
	above_length += sqrt( CGAL::squared_distance(top->point(),newv->point()));
	previousv = top;
	top = newv;
	++above_edges;
      }else{
	below_length += sqrt( CGAL::squared_distance(bot->point(),newv->point()));
	below_edges++;
	previousv = bot;
	bot = newv;
      }
    }


    

    Dijkstra_queue queue;
    source->length=0;
    queue.insert(source);
    while (true) {
      double l;
      Dijkstra_queue::iterator vv=queue.lower_bound(source);
      Delaunay::Vertex_handle v=*vv;
      //      for(Dijkstra_queue::iterator it=queue.begin();it!=queue.end();++it) std::cout<<(*it)->length<<"; ";
      //std::cout<<"       min = "<<v->length<<std::endl;
      //std::cout<<v->point()<<"        "<<v->length<<std::endl;
      queue.erase( vv );
      if( v==target) break;

      Delaunay::Vertex_circulator nn(v),done(nn);
      do{
	l = sqrt( CGAL::squared_distance( v->point(), nn->point()));
	//std::cout<<"        "<<nn->point()<<"        "<<nn->length<<" <? "<<v->length+l<<std::endl;
	if (nn->length > v->length + l ) {
	  if (nn->length<1000000) queue.erase( queue.find( nn ) );
	  nn->length =  v->length + l;
	  nn->previousS = v;
	  queue.insert(nn);
	}
	++nn;
      }while(nn!= done);
    }
    shortest_length=target->length;
    do{
      ++shortest_edges;
      if(target->point().y() > ymax) ymax = target->point().y();
      if(target->point().y() < ymin) ymin = target->point().y();
      target->length=0;
      target=target->previousS;
    }while (target!=source);



    Delaunay::Vertex_circulator vc(source),done(vc);
    do{
      origin_degree++;
      origin_length += sqrt( CGAL::squared_distance( source->point(), vc->point()));
      ++vc;
    }while (vc!=done);




    if (Tmax < ymax) Tmax = ymax;
    if (Tmax < -ymin) Tmax = -ymin;
 
    Torigin_length += origin_length;
    TTorigin_length += origin_length*origin_length;
    Torigin_degree += origin_degree;
    TTorigin_degree += origin_degree*origin_degree;
    Tinternal_length += internal_length;
    TTinternal_length += internal_length*internal_length;
    Tabove_length += above_length;
    TTabove_length += above_length*above_length;
    Tbelow_length += below_length;
    TTbelow_length += below_length*below_length;
    Tslope_length += slope_length;
    TTslope_length += slope_length*slope_length;
    Tshortest_length += shortest_length;
    TTshortest_length += shortest_length*shortest_length;
    Tinternal_edges += internal_edges;
    TTinternal_edges += internal_edges*internal_edges;
    Tabove_edges += above_edges;
    TTabove_edges += above_edges*above_edges;
    Tbelow_edges += below_edges;
    TTbelow_edges += below_edges*below_edges;
    Tslope_edges += slope_edges;
    TTslope_edges += slope_edges*slope_edges;
    Tshortest_edges += shortest_edges;
    TTshortest_edges += shortest_edges*shortest_edges;
    N_it++;


    /* 
    std::cout<<std::endl<<" Trial nb "<<N_it<<std::endl<<
      "internal length (#) | above length (#)    | below length (#)    | min-slope length (#)  | shortest length (#)"<<std::endl<<
      internal_length/LENGTH<<" ("<<internal_edges<<")       | "<<
      above_length/LENGTH   <<" ("<<above_edges   <<")       | "<<
      below_length/LENGTH   <<" ("<<below_edges   <<")       | "<<
      slope_length/LENGTH   <<" ("<<slope_edges   <<")       | "<<
      shortest_length/LENGTH<<" ("<<shortest_edges<<") y in [ "<<ymin/LENGTH<<" , "<<ymax/LENGTH<<" ]"<<
      std::endl;


    std::cout<<" average on "<<N_it<<" trials."<<std::endl<<
      Tinternal_length/N_it/LENGTH<<" ("<<Tinternal_edges/(double)N_it<<")       | "<<
      Tabove_length/N_it/LENGTH   <<" ("<<Tabove_edges/(double)N_it   <<")       | "<<
      Tbelow_length/N_it/LENGTH   <<" ("<<Tbelow_edges/(double)N_it   <<")       | "<<
      Tslope_length/N_it/LENGTH   <<" ("<<Tslope_edges/(double)N_it   <<")       | "<<
      Tshortest_length/N_it/LENGTH<<" ("<<Tshortest_edges/(double)N_it<<")   max y "<<Tmax/LENGTH<<
      std::endl;
    std::cout<<" standard deviation "<<std::endl<<
      sqrt(TTinternal_length/N_it-Tinternal_length*Tinternal_length/N_it/N_it) /LENGTH<<" ("<<
      sqrt(TTinternal_edges/(double)N_it-Tinternal_edges*Tinternal_edges/(double)N_it/(double)N_it)<<")        | "<<
      sqrt(TTabove_length/N_it-Tabove_length*Tabove_length/N_it/N_it) /LENGTH<<" ("<<
      sqrt(TTabove_edges/(double)N_it-Tabove_edges*Tabove_edges/(double)N_it/(double)N_it)         <<")        | "<<
      sqrt(TTbelow_length/N_it-Tbelow_length*Tbelow_length/N_it/N_it) /LENGTH<<" ("<<
      sqrt(TTbelow_edges/(double)N_it-Tbelow_edges*Tbelow_edges/(double)N_it/(double)N_it)         <<")        | "<<
      sqrt(TTslope_length/N_it-Tslope_length*Tslope_length/N_it/N_it) /LENGTH<<" ("<<
      sqrt(TTslope_edges/(double)N_it-Tslope_edges*Tslope_edges/(double)N_it/(double)N_it)         <<")        | "<<
      sqrt(TTshortest_length/N_it-Tshortest_length*Tshortest_length/N_it/N_it) /LENGTH<<" ("<<
      sqrt(TTshortest_edges/(double)N_it-Tshortest_edges*Tshortest_edges/(double)N_it/(double)N_it)<<")        | "<<
      std::endl;
      */std::cout<<"\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"
	       <<N_it<<" th  trial"<<std::flush;


 }

 std::cout<<std::endl<<"%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%"<<std::endl;
 std::cout<<"Averaging on "<<N_trials<<" trials of random points with density "<<N_points<<std::endl
	   <<" we get"<<std::endl<<std::endl
	   <<"\\begin{tabular}{|c|l|l|c|c|}\\hline"<<std::endl
	   <<"path $P$ & $\\Ex{\\length{P}|}$ & $\\sigma(\\length{P})$ & $\\Ex{\\size{P}}/\\sqrt{n}$ & $\\sigma(\\size{P})/\\sqrt{n}$\\\\\\hline"<<std::endl
	   <<"$Star$ &   "
	   <<Torigin_length/(double)N_trials*sqrt(N_points)/LENGTH<<" & "
	  <<sqrt(TTorigin_length/(double)N_trials-Torigin_length*Torigin_length/(double)N_trials/(double)N_trials)*sqrt(N_points)/LENGTH<<"& "
	   <<Torigin_degree/(double)N_trials<<" & "
	              <<sqrt(TTorigin_degree/(double)N_trials-Torigin_degree*Torigin_degree/(double)N_trials/(double)N_trials)<<" \\\\\\hline "<<std::endl
	   <<"$SW$ &  &  & "
	   <<Tinternal_edges/(double)N_trials/sqrt(N_points)<<" & "
	              <<sqrt(TTinternal_edges/(double)N_trials-Tinternal_edges*Tinternal_edges/(double)N_trials/(double)N_trials)/sqrt(N_points)<<" \\\\\\hline "<<std::endl
	   <<"$UP$ & "<<Tabove_length/N_trials/LENGTH<<" & "
	              <<sqrt(TTabove_length/N_trials-Tabove_length*Tabove_length/N_trials/N_trials) /LENGTH<<" & "
	   <<Tabove_edges/(double)N_trials/sqrt(N_points)<<" & "
	              <<sqrt(TTabove_edges/(double)N_trials-Tabove_edges*Tabove_edges/(double)N_trials/(double)N_trials)/sqrt(N_points)<<" \\\\\\hline "<<std::endl
	   <<"$GP$ & "<< Tslope_length/N_trials/LENGTH<<" & "
	              <<sqrt(TTslope_length/N_trials-Tslope_length*Tslope_length/N_trials/N_trials) /LENGTH<<" & "
	              <<Tslope_edges/(double)N_trials/sqrt(N_points)<<" & "
	              <<sqrt(TTslope_edges/(double)N_trials-Tslope_edges*Tslope_edges/(double)N_trials/(double)N_trials)/sqrt(N_points) <<" \\\\\\hline "<<std::endl
	   <<"$SP$ & "<<Tshortest_length/N_trials/LENGTH <<" & "
	   << sqrt(TTshortest_length/N_trials-Tshortest_length*Tshortest_length/N_trials/N_trials) /LENGTH<<" & "
	   <<Tshortest_edges/(double)N_trials/sqrt(N_points)<<" & "
	   <<sqrt(TTshortest_edges/(double)N_trials-Tshortest_edges*Tshortest_edges/(double)N_trials/(double)N_trials)/sqrt(N_points) <<" \\\\\\hline "<<std::endl
	   <<"\\end{tabular}\\\\ "<<std::endl;
  std::cout<<"%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%"<<std::endl;



}
